local json = require("lib/utils/accelbyte/json")

local husk_names = {
	wild = "units/pd2_dlc_wild/characters/npc_criminals_wild_1/player_criminal_wild_husk",
	joy = "units/pd2_dlc_joy/characters/npc_criminals_joy_1/player_criminal_joy_husk"
}

local function get_husk_unit(character_id)
	return husk_names[character_id] or tweak_data.blackmarket.characters[character_id].npc_unit:gsub("(.+)/npc_", "%1/player_") .. "_husk"
end

local function spawn_husk_unit(parent_unit)
	local character_id = managers.network:session():local_peer():character_id()
	local husk_unit = get_husk_unit(character_id)
	local husk_unit_idstring = Idstring(husk_unit)

	if not DB:has(Idstring("unit"), husk_unit_idstring) then
		return
	end

	local spawned_husk_unit = World:spawn_unit(husk_unit_idstring, parent_unit:position(), Rotation())
	Network:detach_unit(spawned_husk_unit)

	local husk_base = spawned_husk_unit:base()
	local husk_movement = spawned_husk_unit:movement()
	local husk_inventory = spawned_husk_unit:inventory()
	local husk_sound = spawned_husk_unit:sound()
	local husk_damage = spawned_husk_unit:damage()

	husk_base.pre_destroy = function(self, unit)
		unit:movement():pre_destroy(unit)
		unit:inventory():pre_destroy(unit)
		UnitBase.pre_destroy(self, unit)
	end

	spawned_husk_unit:contour().add = function() end

	husk_movement._parent_unit = parent_unit

	husk_movement.update = function(self, ...) end

	husk_movement._register_revive_SO = function() end
	husk_movement.set_need_assistance = function(self, need_assistance) self._need_assistance = need_assistance end
	husk_movement.set_need_revive = function(self, need_revive) self._need_revive = need_revive end

	husk_movement.update_visual_state = function(self)
		local player_peer = managers.network:session():local_peer()
		local character = player_peer:character()

		local character_table = managers.criminals:characters()
		character_table[0] = {
			peer_id = false,
			taken = true,
			name = character,
			unit = self._unit,
			static_data = {},
			data = {},
			visual_state = {}
		}

		local old_peer_unit = player_peer._unit
		local old_peer_id_function = player_peer.id

		player_peer._unit = self._unit
		player_peer.id = function() return -1 end

		player_peer:update_character_visual_state()

		player_peer._unit = old_peer_unit
		player_peer.id = old_peer_id_function

		character_table[0] = nil

		self._unit:inventory():set_mask_visibility(true, player_peer:blackmarket_outfit().mask.mask_id, player_peer:mask_blueprint())
	end

	husk_inventory.set_mask_visibility = function(self, state, mask_id, mask_blueprint)
		self._mask_id = mask_id or self._mask_id
		self._mask_blueprint = mask_blueprint or self._mask_blueprint
		if not self._mask_id or not self._mask_blueprint then return end

		self._mask_visibility = state

		if alive(self._mask_unit) then
			if not state then
				for _, linked_unit in ipairs(self._mask_unit:children()) do
					linked_unit:unlink()
					World:delete_unit(linked_unit)
				end

				self._mask_unit:unlink()

				local name = self._mask_unit:name()

				World:delete_unit(self._mask_unit)
			end

			return
		end

		if not state then
			return
		end

		local mask_unit_name = managers.blackmarket:mask_unit_name_by_mask_id(self._mask_id)

		if not managers.dyn_resource:is_resource_ready(Idstring("unit"), mask_unit_name, managers.dyn_resource.DYN_RESOURCES_PACKAGE) then
			return
		end

		mask_unit_name = mask_unit_name[Global.level_data.level_id] or mask_unit_name.default or mask_unit_name
		local mask_align = self._unit:get_object(Idstring("Head"))
		local mask_unit = World:spawn_unit(Idstring(mask_unit_name), mask_align:position(), mask_align:rotation())

		mask_unit:base():apply_blueprint(self._mask_blueprint)
		self._unit:link(mask_align:name(), mask_unit)

		self._mask_unit = mask_unit
		local peer = managers.network:session():peer_by_unit(self._unit)
		local mask_data = {
			mask_id = self._mask_id,
			mask_unit = mask_unit,
			mask_align = mask_align,
			character_name = character_id
		}

		self:update_mask_offset(mask_data)

		if not self._mask_id or not tweak_data.blackmarket.masks[self._mask_id].type then
			local backside = World:spawn_unit(Idstring("units/payday2/masks/msk_backside/msk_backside"), mask_align:position(), mask_align:rotation())

			self._mask_unit:link(self._mask_unit:orientation_object():name(), backside, backside:orientation_object():name())
		end

		self._unit:damage():run_sequence_simple(managers.blackmarket:character_mask_off_sequence_by_character_name(character_id))

		if not self._mask_id or not tweak_data.blackmarket.masks[self._mask_id].skip_mask_on_sequence then
			local mask_on_sequence = managers.blackmarket:character_mask_on_sequence_by_character_name(character_id)

			if mask_on_sequence then
				--self._unit:damage():run_sequence_simple(mask_on_sequence)
			end
		end
	end

	husk_sound.say = function () end
	husk_sound.play = function () end
	husk_sound._play = function () end

	husk_damage:run_sequence_simple(managers.blackmarket:character_sequence_by_character_name(character_id))

	husk_movement:set_character_anim_variables()
	husk_movement:update_visual_state()

	local level_data = managers.job and managers.job:current_level_data()
	if level_data and level_data.player_sequence then
		husk_damage:run_sequence_simple(level_data.player_sequence)
	end

	managers.groupai:state():unregister_criminal(spawned_husk_unit)

	return spawned_husk_unit
end

KaraokeManager = KaraokeManager or class()

function KaraokeManager:init()
	self._fullscreen_workspace = managers.gui_data:create_fullscreen_workspace()
	self._saferect_workspace = managers.gui_data:create_saferect_workspace()

	self._fullscreen_panel = self._fullscreen_workspace:panel()
	self._panel = self._saferect_workspace:panel()

	self._karaoke_audio_panel = KaraokeAudioGui:new(self._fullscreen_panel, {}, 0, 0, 0, 0)

	self._karaoke_menu_panel = KaraokeMenuGui:new(self._panel, {}, 0, 0, self._panel:width(), self._panel:height())
	self._karaoke_score_panel = KaraokeScoreGui:new(self._panel, {}, 0, 0, self._panel:width(), self._panel:height())

	self._karaoke_splash_panel = KaraokeSplashGui:new(self._fullscreen_panel, {}, 0, 0, self._fullscreen_panel:width(), self._fullscreen_panel:height())
	self._karaoke_panel = KaraokeGui:new(self._panel, {}, 0, 0, self._panel:width(), self._panel:height())

	self._karaoke_menu_panel:hide()
	self._karaoke_score_panel:hide()
	self._karaoke_splash_panel:hide()
	self._karaoke_panel:hide()

	self._fade_rect = self._fullscreen_panel:rect({
		w = self._fullscreen_panel:width(),
		h = self._fullscreen_panel:height(),
		x = 0,
		y = 0,
		color = Color.white,
		layer = 3
	})

	self._fade_rect:set_alpha(0)

	self:hide_workspace()
	self:show_workspace()

	self:_setup_controller({
		keyboard = {
			triangle = "i",
			square   = "j",
			cross    = "k",
			circle   = "l",

			accept = "enter",
			cancel = "esc",

			-- debug_rewind = "z",
			-- debug_pause = "x",
			-- debug_forward = "c",
			-- debug_restart = "v"
		},
		xbox360pad = {
			extra_up    = "d_up",
			extra_down  = "d_down",
			extra_left  = "d_left",
			extra_right = "d_right",

			triangle = "y",
			square   = "x",
			cross    = "a",
			circle   = "b",

			accept = "a",
			cancel = "b"
		}
	})

	self:_clean_inputs()

	self._line_index = 1
	self._note_index = 1

	self._active_lines = {}
	self._active_note_index = 1
	self._active_notes = {}

	self._score_data = {
		great = 0,
		good = 0,
		missed = 0
	}

	self._all_song_data = {}
	self._current_song_data = {}

	self:_load_songs()

	for index, song in pairs(self._all_song_data) do
		self._karaoke_menu_panel:set_song_info(index, song.settings.title, song.settings.disc_sticker)
	end

	self._song_select_index = 1
	self._karaoke_menu_panel:set_selection(self._song_select_index)

	self._karaoke_position = Vector3()
	self._karaoke_rotation = Rotation()

	if Global.game_settings.level_id == "chill" then
		World:spawn_unit(Idstring("units/pd2_mod_karaoke/equipment/gen_interactable_karaoke_machine/gen_interactable_karaoke_machine"), Vector3(1838, 744, -0.76), Rotation(-90, 0, 0))
	end

	-- self:open()
end

function KaraokeManager:active()
	return self._current_state ~= nil
end

function KaraokeManager:trigger_fade()
	function animate_fade(o)
		local in_time = 4
		local out_time = 1

		local in_lerp_speed = 1 / in_time
		local out_lerp_speed = 1 / out_time

		local t, dt = 0
		local lerp = 0

		local left_lerped = start_left

		self._up = true

		while true do
			dt = coroutine.yield()
			t = t + dt

			if self._up then
				lerp = math.min(t * in_lerp_speed, 1)

				self._fade_rect:set_alpha(lerp)

				if lerp >= 1 then
					t = 0
					self._up = false

					self._fade_halfway = true
				end
			else
				lerp = 1 - math.min(t * out_lerp_speed, 1)

				self._fade_rect:set_alpha(lerp)

				if lerp <= 0 then
					self._fade_halfway = false
					return
				end
			end
		end
	end

	self._fade_rect:animate(animate_fade)
end

function KaraokeManager:fade_halfway()
	return not not self._fade_halfway
end

function KaraokeManager:open(position, rotation)
	self._karaoke_splash_panel = KaraokeSplashGui:new(self._fullscreen_panel, {}, 0, 0, self._fullscreen_panel:width(), self._fullscreen_panel:height())
	self._fade_rect = self._fullscreen_panel:rect({
		w = self._fullscreen_panel:width(),
		h = self._fullscreen_panel:height(),
		x = 0,
		y = 0,
		color = Color.white,
		layer = 3
	})

	self._karaoke_splash_panel:hide()
	self._fade_rect:set_alpha(0)

	SoundDevice:set_rtpc("option_music_volume", 0)

	self._karaoke_position = position or Vector3()
	self._karaoke_rotation = rotation or Vector3()

	managers.hud:set_disabled()

	self._controller:set_enabled(true)

	self._karaoke_menu_panel:show()

	self._current_state = "menu"
end

function KaraokeManager:close()
	SoundDevice:set_rtpc("option_music_volume", Global.music_manager.volume * 100)

	managers.hud:set_enabled()

	self._controller:set_enabled(false)

	self._karaoke_score_panel:hide()
	self._karaoke_menu_panel:hide()
	self._karaoke_panel:hide()
	self._karaoke_splash_panel:hide()

	self._karaoke_audio_panel:pause_all()

	self._current_state = nil
end

function KaraokeManager:goto_score(t)
	self._karaoke_menu_panel:hide()
	self._karaoke_panel:hide()
	self._karaoke_splash_panel:hide()
	self._karaoke_score_panel:hide()

	self._score_start_time = t

	self._current_state = "score"
end

local karaoke_folder = ModPath .. "/songs/"
function KaraokeManager:_load_songs()
	self._karaoke_folder = karaoke_folder

	local potential_songs = SystemFS:list(self._karaoke_folder)
	if potential_songs then
		for _, potential_song in pairs(potential_songs) do
			if string.ends(potential_song, ".json") then
				local file = io.open(self._karaoke_folder .. potential_song, "r")
				local song_data = json.decode(file:read("*all"))

				table.insert(self._all_song_data, song_data)
			end
		end
	end
end

function KaraokeManager:_setup_controller(custom_inputs)
	self._controller = managers.controller:create_controller("karaoke_controller", managers.controller:get_default_wrapper_index(), false, 35)

	self._custom_controller_inputs = {}

	for controller_id, inputs in pairs(custom_inputs) do
		for input_id, input_button in pairs(inputs) do
			self._custom_controller_inputs[input_id] = self._custom_controller_inputs[input_id] or {}
			self._custom_controller_inputs[input_id][controller_id] = input_button
		end
	end
end

function KaraokeManager:_get_input(mode, input, override_controller_id)
	if not self._custom_controller_inputs[input] then return false end

	local controller_id = override_controller_id or self._controller:get_default_controller_id()
	if not self._custom_controller_inputs[input][controller_id] then return false end

	local controller = self._controller:get_controller(controller_id)
	if not (controller and controller[mode]) then return false end

	local button_id = self._custom_controller_inputs[input][controller_id]
	return controller[mode](controller, Idstring(button_id))
end

function KaraokeManager:start_song(t)
	self._current_song_data = self._all_song_data[self._song_select_index]

	self._line_index = 1
	self._note_index = 1

	self._active_lines = {}
	self._active_note_index = 1
	self._active_notes = {}

	self._score_data = {
		great = 0,
		good = 0,
		missed = 0
	}

	self._karaoke_menu_panel:hide()

	self._karaoke_splash_panel:set_splash_texture(self._current_song_data.settings.splash)

	self._karaoke_audio_panel:set_audio_movie(self._current_song_data.settings.audio_tracks.vocals, false, 1)
	self._karaoke_audio_panel:set_audio_movie(self._current_song_data.settings.audio_tracks.instrumentals, false, 2)

	local camera_ids = Idstring("units/pd2_mod_karaoke/cutscene_camera/cutscene_camera_lower_priority")
	local camera_unit = World:spawn_unit(camera_ids, self._karaoke_position, self._karaoke_rotation)

	self._camera_unit = camera_unit

	local camera_asm = camera_unit:anim_state_machine()
	camera_asm:play_raw(Idstring("anims/karaoke/player_camera/main"), {
		blend_set = "all",
		type = "loop"
	})

	self._husk_unit = spawn_husk_unit(camera_unit)
	self._husk_unit:set_position(self._karaoke_position)
	self._husk_unit:set_rotation(self._karaoke_rotation)
	self._husk_unit:anim_state_machine():set_enabled(true)
	self._husk_unit:set_driving("animation")

	self._husk_unit:anim_state_machine():play_raw(Idstring("anims/karaoke/player/idle"), {
		blend_set = "all",
		type = "loop"
	})

	self._mic_unit = World:spawn_unit(Idstring("units/pd2_mod_karaoke/props/microphone/microphone"), Vector3())
	self._husk_unit:link(Idstring("RightHand"), self._mic_unit, self._mic_unit:orientation_object():name())

	self:set_player_visibility(false)

	self._song_start_time = t
	self._current_state = "song"
end

function KaraokeManager:set_player_visibility(state)
	local player_unit = managers.player:player_unit()
	if player_unit and alive(player_unit) then
		local camera_unit = player_unit:camera():camera_unit()
		if camera_unit and alive(camera_unit) then
			camera_unit:set_visible(state)

			for unit_id, unit_entry in pairs(camera_unit:spawn_manager():spawned_units()) do
				if alive(unit_entry.unit) then
					unit_entry.unit:set_visible(state)
				end
			end

			if state then
				camera_unit:base():show_weapon()
			else
				camera_unit:base():hide_weapon()
			end
		end
	end
end

function KaraokeManager:end_song()
	if self._in_cutscene then
		self:handle_cutscene_stop()
	end

	self._cutscene_done = false
	self._cutscene_available = false

	if self._camera_unit and alive(self._camera_unit) then
		World:delete_unit(self._camera_unit)
	end

	if self._husk_unit and alive(self._husk_unit) then
		World:delete_unit(self._husk_unit)
	end

	if self._mic_unit and alive(self._mic_unit) then
		World:delete_unit(self._mic_unit)
	end

	self:set_player_visibility(true)

	self._current_state = nil
end

function KaraokeManager:paused_update(t, dt)
	self:update(Application:time(), 0.1, true)
end

function KaraokeManager:update(t, dt, paused)
	-- local player_unit = managers.player:player_unit()
	-- if player_unit and alive(player_unit) then
	-- 	log(tostring(player_unit:position()))
	-- end

	if paused and not self._paused then
		self._paused = true

		self._karaoke_audio_panel:pause_all()
	elseif not paused and self._paused then
		self._paused = false

		self._karaoke_audio_panel:play_all()
	end

	if self._paused then return end
	if not self._current_state then return end

	self:_update_input(t, dt)

	if self._current_state == "song" then
		self:_update_song(t, dt)
	elseif self._current_state == "menu" then
		self:_update_menu(t, dt)
	elseif self._current_state == "score" then
		self:_update_score(t, dt)
	end
end

function KaraokeManager:_clean_inputs()
	self._input = {
		up = {
			pressed = false,
			down = false
		},
		down = {
			pressed = false,
			down = false
		},
		right = {
			pressed = false,
			down = false
		},
		left = {
			pressed = false,
			down = false
		},
		triangle = {
			pressed = false,
			down = false
		},
		square = {
			pressed = false,
			down = false
		},
		circle = {
			pressed = false,
			down = false
		},
		cross = {
			pressed = false,
			down = false
		},
		accept = {
			pressed = false,
			down = false
		},
		cancel = {
			pressed = false,
			down = false
		}
	}
end

local DEADZONE = 0.1
function KaraokeManager:_update_input(t, dt)
	if self._controller_enabled ~= self._controller:enabled() then
		if self._controller_enabled then
			self._controller_enabled = self._controller:enabled()

			self:_clean_inputs()

			return
		end
	elseif not self._controller_enabled then
		self:_clean_inputs()

		return
	end

	self._controller_enabled = self._controller:enabled()

	-- Directional inputs!
	-- Convert axis input to button-like input!
	local move = self._controller:get_input_axis("move")

	local move_y = mvector3.y(move)
	local move_x = mvector3.x(move)

	local up_down    = (move_y >  DEADZONE) or self:_get_input("down", "extra_up")
	local down_down  = (move_y < -DEADZONE) or self:_get_input("down", "extra_down")
	local right_down = (move_x >  DEADZONE) or self:_get_input("down", "extra_right")
	local left_down  = (move_x < -DEADZONE) or self:_get_input("down", "extra_left")

	-- Also factor in some extra bindings so that the d-pad and the joystick work.
	local up_pressed    = (not self._input.up.down    and up_down   ) or self:_get_input("pressed", "extra_up")
	local down_pressed  = (not self._input.down.down  and down_down ) or self:_get_input("pressed", "extra_down")
	local right_pressed = (not self._input.right.down and right_down) or self:_get_input("pressed", "extra_right")
	local left_pressed  = (not self._input.left.down  and left_down ) or self:_get_input("pressed", "extra_left")

	-- Note inputs!
	self._input = {
		up = {
			pressed = up_pressed,
			down = up_down
		},
		down = {
			pressed = down_pressed,
			down = down_down
		},
		right = {
			pressed = right_pressed,
			down = right_down
		},
		left = {
			pressed = left_pressed,
			down = left_down
		},
		triangle = {
			pressed = self:_get_input("pressed", "triangle"),
			down = self:_get_input("down", "triangle")
		},
		square = {
			pressed = self:_get_input("pressed", "square"),
			down = self:_get_input("down", "square")
		},
		circle = {
			pressed = self:_get_input("pressed", "circle"),
			down = self:_get_input("down", "circle")
		},
		cross = {
			pressed = self:_get_input("pressed", "cross"),
			down = self:_get_input("down", "cross")
		},
		accept = {
			pressed = self:_get_input("pressed", "accept"),
			down = self:_get_input("down", "accept"),
		},
		cancel = {
			pressed = self:_get_input("pressed", "cancel"),
			down = self:_get_input("down", "cancel"),
		}
	}

	if not self._debug_forward then
		if self:_get_input("down", "debug_rewind") then
			if not self._debug_rewinding then
				self._debug_rewinding = true
				self._karaoke_audio_panel:pause_all()

				self._debug_paused = true
				self._paused = true
			end

			local time = self._karaoke_audio_panel:audio_time()
			local new_time = time - (60 * dt)
			if new_time < 0 then
				new_time = 0
			end

			self._karaoke_audio_panel:goto_time(1, new_time)
			self._karaoke_audio_panel:goto_time(2, new_time)

			self._karaoke_panel:hide_timer_bar(1)
			self._karaoke_panel:hide_timer_bar(2)
			self._karaoke_panel:hide_timer_bar(3)
			self._karaoke_panel:hide_timer_bar(4)

			self._line_index = 1
			self._note_index = 1

			self._active_lines = {}

			self._active_note_index = 1
			self._active_notes = {}
		elseif self._debug_rewinding then
			self._debug_rewinding = false
		end
	end

	if not self._debug_rewinding then
		if self:_get_input("down", "debug_forward") then
			if not self._debug_forward then
				self._debug_forward = true
				self._karaoke_audio_panel:pause_all()

				self._debug_paused = true
				self._paused = true
			end

			local time = self._karaoke_audio_panel:audio_time()
			local new_time = time + (60 * dt)
			if new_time < 0 then
				new_time = 0
			end

			self._karaoke_audio_panel:goto_time(1, new_time)
			self._karaoke_audio_panel:goto_time(2, new_time)
		elseif self._debug_forward then
			self._debug_forward = false
		end
	end

	if not (self._debug_rewinding or self._debug_forward) then
		if self:_get_input("pressed", "debug_pause") then
			self._debug_paused = not self._debug_paused
		end

		if self._debug_paused and not self._paused then
			self._paused = true
			self._karaoke_audio_panel:pause_all()
		elseif not self._debug_paused and self._paused then
			self._paused = false
			self._karaoke_audio_panel:play_all()
		end

		if self:_get_input("pressed", "debug_restart") then
			self._karaoke_audio_panel:goto_time(1, 0)
			self._karaoke_audio_panel:goto_time(2, 0)

			self._karaoke_panel:hide_timer_bar(1)
			self._karaoke_panel:hide_timer_bar(2)
			self._karaoke_panel:hide_timer_bar(3)
			self._karaoke_panel:hide_timer_bar(4)

			self._line_index = 1
			self._note_index = 1

			self._active_lines = {}

			self._active_note_index = 1
			self._active_notes = {}
		end
	end
end

function KaraokeManager:_update_song(t, dt)
	self:_update_splash(t, dt)
	if self._splash_state then return end

	self:_update_song_time(t, dt)

	self:_update_next_lines_notes()
	self:_update_active_lines()
	self:_update_active_notes()

	self._karaoke_panel:check_debug_times()

	local hit_notes = self._score_data.good + self._score_data.great
	if hit_notes ~= self._last_hit_notes then
		if self._husk_unit and alive(self._husk_unit) then
			if hit_notes == 2 then
				self._husk_unit:anim_state_machine():play_raw(Idstring("anims/karaoke/player/normal"), {
					blend_set = "all",
					type = "loop"
				})
			elseif hit_notes == 65 then
				self._husk_unit:anim_state_machine():play_raw(Idstring("anims/karaoke/player/good"), {
					blend_set = "all",
					type = "loop"
				})
			end
		end

		if self._current_song_data.settings.notes_for_cutscene and self._current_song_data.settings.cutscene_start and hit_notes == self._current_song_data.settings.notes_for_cutscene then
			self._cutscene_available = true
		end
	end

	if not self._cutscene_available and self._extrapolated_song_time >= (self._current_song_data.settings.cutscene_start - 4) then
		self._cutscene_done = true
	end

	self._last_hit_notes = hit_notes

	if not self._cutscene_done then
		if not self._in_cutscene and not self._triggered_cutscene_fade and self._cutscene_available and self._extrapolated_song_time >= (self._current_song_data.settings.cutscene_start - 4) then
			self:trigger_fade()

			self._triggered_cutscene_fade = true
		end

		if not self._in_cutscene and self._triggered_cutscene_fade and self._extrapolated_song_time >= self._current_song_data.settings.cutscene_start then
			self:handle_cutscene()

			self._triggered_cutscene_fade = false
			self._in_cutscene = true
		end

		if self._in_cutscene and not self._triggered_cutscene_fade and self._cutscene_available and self._extrapolated_song_time >= (self._current_song_data.settings.cutscene_end - 4) then
			self:trigger_fade()

			self._triggered_cutscene_fade = true
		end

		if self._in_cutscene and self._triggered_cutscene_fade and self._extrapolated_song_time >= self._current_song_data.settings.cutscene_end then
			self:handle_cutscene_stop()
		end
	end

	if not self._triggered_exit_fade and self._extrapolated_song_time >= (self._karaoke_audio_panel:length(1) - 4) then
		self:trigger_fade()

		self._triggered_exit_fade = true
	end

	if self._karaoke_audio_panel:all_done() then
		self:end_song()
		self:goto_score(t)

		self._triggered_exit_fade = false
		self._last_hit_notes = nil
	end
end

function KaraokeManager:handle_cutscene()
	if self._current_song_data.settings.cutscene_pre_rendered then
		self._pre_rendered_cutscene_bg = self._fullscreen_panel:rect({
			w = self._fullscreen_panel:width(),
			h = self._fullscreen_panel:height(),
			x = 0,
			y = 0,
			color = Color.black,
			layer = 1
		})

		self._pre_rendered_cutscene = self._fullscreen_panel:video({
			video = self._current_song_data.settings.cutscene_pre_rendered,
			loop = true,
			layer = 2
		})
		managers.video:add_video(self._pre_rendered_cutscene)

		local fullscreen_panel_width = self._fullscreen_panel:width()
		local fullscreen_panel_height = self._fullscreen_panel:height()

		local video_aspect = self._pre_rendered_cutscene:video_width()/self._pre_rendered_cutscene:video_height()
		local fullscreen_panel_aspect = fullscreen_panel_width/fullscreen_panel_height

		local video_width = fullscreen_panel_width
		local video_height = fullscreen_panel_height

		if video_aspect < fullscreen_panel_aspect then
			video_width = fullscreen_panel_height * video_aspect
		else
			video_height = fullscreen_panel_width / video_aspect
		end

		self._pre_rendered_cutscene:set_size(video_width, video_height)
		self._pre_rendered_cutscene:set_center(fullscreen_panel_width / 2, fullscreen_panel_height / 2)

		return
	end

	if self._current_song_data.settings.cutscene_animation and self._current_song_data.settings.cutscene_camera then
		self._cached_husk_position = self._husk_unit:position()
		self._husk_unit:set_position(Vector3(10000, 10000, 10000))

		local position = self._karaoke_position
		if self._current_song_data.settings.cutscene_position then
			position = Vector3(
				self._current_song_data.settings.cutscene_position.x,
				self._current_song_data.settings.cutscene_position.y,
				self._current_song_data.settings.cutscene_position.z
			)
		end

		local rotation = self._karaoke_rotation
		if self._current_song_data.settings.cutscene_rotation then
			rotation = Rotation(
				self._current_song_data.settings.cutscene_rotation.yaw,
				self._current_song_data.settings.cutscene_rotation.pitch,
				self._current_song_data.settings.cutscene_rotation.roll
			)
		end

		local camera_ids = Idstring("units/pd2_mod_karaoke/cutscene_camera/cutscene_camera")
		local camera_unit = World:spawn_unit(camera_ids, position, rotation)

		self._cutscene_camera_unit = camera_unit

		local camera_asm = camera_unit:anim_state_machine()
		camera_asm:play_raw(Idstring(self._current_song_data.settings.cutscene_camera), {
			blend_set = "all",
			type = "once"
		})

		self._cutscene_husk_unit = spawn_husk_unit(camera_unit)
		self._cutscene_husk_unit:set_position(position)
		self._cutscene_husk_unit:set_rotation(rotation)
		self._cutscene_husk_unit:anim_state_machine():set_enabled(true)
		self._cutscene_husk_unit:set_driving("animation")

		self._cutscene_husk_unit:anim_state_machine():play_raw(Idstring(self._current_song_data.settings.cutscene_animation), {
			blend_set = "all",
			type = "once"
		})

		-- self._mic_unit = World:spawn_unit(Idstring("units/pd2_mod_karaoke/props/microphone/microphone"), Vector3())
		-- self._husk_unit:link(Idstring("RightHand"), self._mic_unit, self._mic_unit:orientation_object():name())
	end
end

function KaraokeManager:handle_cutscene_stop()
	self._triggered_cutscene_fade = false
	self._in_cutscene = false

	self._cutscene_done = true

	if self._pre_rendered_cutscene and alive(self._pre_rendered_cutscene) and self._pre_rendered_cutscene_bg and alive(self._pre_rendered_cutscene_bg) then
		managers.video:remove_video(self._pre_rendered_cutscene)

		self._fullscreen_panel:remove(self._pre_rendered_cutscene_bg)
		self._fullscreen_panel:remove(self._pre_rendered_cutscene)

		self._pre_rendered_cutscene = nil
	end

	if self._current_song_data.settings.cutscene_animation and self._current_song_data.settings.cutscene_camera then
		self._husk_unit:set_position(self._cached_husk_position)
		self._cached_husk_position = nil

		if self._cutscene_husk_unit and alive(self._cutscene_husk_unit) then
			World:delete_unit(self._cutscene_husk_unit)
		end

		if self._cutscene_camera_unit and alive(self._cutscene_camera_unit) then
			World:delete_unit(self._cutscene_camera_unit)
		end
	end
end

function KaraokeManager:_update_splash(t, dt)
	local time_since_start = t - self._song_start_time

	if time_since_start <= 2.4 and not self._splash_state then
		self._karaoke_splash_panel:show()
		self._karaoke_panel:hide()

		-- self._karaoke_audio_panel:pause_all()

		self._splash_state = true
	elseif time_since_start > 2.4 and self._splash_state then 
		self._karaoke_splash_panel:hide()
		self._karaoke_panel:show()

		self._karaoke_audio_panel:play_all()

		self._splash_state = false
	end
end

function KaraokeManager:_update_song_time(t, dt)
	local next_song_time = self._karaoke_audio_panel:audio_time(1)
	if self._absolute_song_time ~= next_song_time then
		self._extrapolated_song_time = next_song_time
		self._absolute_song_time = next_song_time
	end

	if not self._paused then
		self._extrapolated_song_time = self._extrapolated_song_time + dt
	end

	self._karaoke_panel:debug_timer(self._extrapolated_song_time)
end

function KaraokeManager:_update_next_lines_notes()
	local next_line = self._current_song_data.lines[self._line_index]
	while next_line and self._extrapolated_song_time >= next_line.spawn_time do
		self._karaoke_panel:set_timer_bar(next_line.row, next_line.lyric_texture, next_line.lyric_width)
		self._karaoke_panel:add_debug_to_line(next_line.row, next_line.start_time, next_line.end_time)

		self._active_lines[next_line.row] = next_line

		local line_length = next_line.end_time - next_line.start_time

		local next_note = self._current_song_data.notes[self._note_index]
		while next_note and next_note.start_time < next_line.end_time and next_note.start_time > next_line.start_time do
			self._karaoke_panel:add_note(
				self._note_index,
				self._controller.TYPE == "pc" and "pc" or "ps",
				next_note.button,
				next_note.type,
				next_line.row,
				(next_note.start_time - next_line.start_time) / line_length,
				((next_note.start_time + next_note.length) - next_line.start_time) / line_length
			)
			self._karaoke_panel:add_debug_to_note(self._note_index, next_note.start_time)
			self._active_notes[self._note_index] = next_note

			self._note_index = self._note_index + 1
			next_note = self._current_song_data.notes[self._note_index]
		end

		self._line_index = self._line_index + 1
		next_line = self._current_song_data.lines[self._line_index]
	end
end

function KaraokeManager:_update_active_lines()
	for row, line in pairs(self._active_lines) do
		if self._extrapolated_song_time >= line.despawn_time then
			self._karaoke_panel:hide_timer_bar(row)
			self._active_lines[row] = nil
		end

		if self._extrapolated_song_time >= line.end_time then
			self._karaoke_panel:set_timer_bar_fill(row, 1)
		elseif self._extrapolated_song_time >= line.start_time then
			self._karaoke_panel:set_marker_bar(row)

			local fill_percentage = (self._extrapolated_song_time - line.start_time) / (line.end_time - line.start_time)
			self._karaoke_panel:set_timer_bar_fill(row, fill_percentage)			
		else
			self._karaoke_panel:set_timer_bar_fill(row, 0)
		end
	end
end

function KaraokeManager:_update_active_notes()
	if self._hold_start_t and self._hold_length then
		local button_down = self._input[self._hold_button].down

		if not button_down or self._extrapolated_song_time >= (self._hold_start_t + self._hold_length) then
			local hold_percentage = math.min((self._extrapolated_song_time - self._hold_start_t) / self._hold_length, 1)

			self._karaoke_audio_panel:set_audio_movie("movies/karaoke_sfx/note_hit", false, 3)
			self._karaoke_audio_panel:play(3)

			if hold_percentage > 0.9 then
				self._score_data["great"] = self._score_data["great"] + 1
			elseif hold_percentage > 0.5 then
				self._score_data["good"] = self._score_data["good"] + 1
			else
				self._score_data["missed"] = self._score_data["missed"] + 1
			end

			self._hold_button = nil
			self._hold_start_t = nil
			self._hold_length = nil
		else
			return
		end
	end

	if self._rapid_press_button and self._rapid_press_total then
		if self._extrapolated_song_time >= self._rapid_press_end_t then
			local percentage = self._rapid_press_count / self._rapid_press_total

			if percentage > 1 then
				self._score_data["great"] = self._score_data["great"] + 1
			elseif percentage > 0.5 then
				self._score_data["good"] = self._score_data["good"] + 1
			else
				self._score_data["missed"] = self._score_data["missed"] + 1
			end

			self._rapid_press_button = nil
			self._rapid_press_total = nil
			self._rapid_press_count = nil
			self._rapid_press_end_t = nil
		else
			if self._input[self._rapid_press_button].pressed then
				self._rapid_press_count = self._rapid_press_count + 1

				self._karaoke_audio_panel:set_audio_movie("movies/karaoke_sfx/note_hit", false, 3)
				self._karaoke_audio_panel:play(3)
			end

			return
		end
	end

	local great_timing = self._current_song_data.settings.great_timing
	local good_timing  = self._current_song_data.settings.good_timing

	local note = self._active_notes[self._active_note_index]
	if not note then return end

	local note_zone = nil

	local good_timing_start = note.start_time - good_timing
	local good_timing_end   = note.start_time + good_timing

	if self._extrapolated_song_time >= good_timing_start and self._extrapolated_song_time <= good_timing_end then
		note_zone = "good"

		local great_timing_start = note.start_time - great_timing
		local great_timing_end   = note.start_time + great_timing

		if self._extrapolated_song_time >= great_timing_start and self._extrapolated_song_time <= great_timing_end then
			note_zone = "great"
		end
	elseif self._extrapolated_song_time > good_timing_end then
		note_zone = "missed"
	end

	if note_zone == "missed" then
		self._active_notes[self._active_note_index] = nil
		self._active_note_index = self._active_note_index + 1

		self._last_missed = true
		self._score_data["missed"] = self._score_data["missed"] + 1
	elseif note_zone and self._input[note.button] and self._input[note.button].pressed then
		local default_effect = note.type == "hold" and "movies/karaoke_sfx/note_hit_long" or "movies/karaoke_sfx/note_hit"
		local sound_effect = note.sound_effect or default_effect

		self._karaoke_audio_panel:set_audio_movie(sound_effect, false, 3)
		self._karaoke_audio_panel:play(3)

		self._karaoke_panel:hit_note(self._active_note_index)

		self._active_notes[self._active_note_index] = nil
		self._active_note_index = self._active_note_index + 1

		self._last_missed = false

		if note.type == "hold" then
			self._hold_button = note.button
			self._hold_start_t = note.start_time
			self._hold_length = note.length
		elseif note.type == "rapid_press" then
			self._rapid_press_button = note.button
			self._rapid_press_total = math.min(5 * note.length, 1)
			self._rapid_press_count = 1
			self._rapid_press_end_t = note.start_time + note.length
		else
			self._score_data[note_zone] = self._score_data[note_zone] + 1
		end
	end

	-- if self._last_missed then
	-- 	self._karaoke_audio_panel:set_volume_gain(1, 0)
	-- else
	-- 	local volume = managers.user:get_setting("sfx_volume")
	-- 	local percentage = (volume - tweak_data.menu.MIN_SFX_VOLUME) / (tweak_data.menu.MAX_SFX_VOLUME - tweak_data.menu.MIN_SFX_VOLUME)

	-- 	self._karaoke_audio_panel:set_volume_gain(1, percentage)
	-- end
end

function KaraokeManager:_update_menu(t, dt)
	if self._input.up.pressed then
		self._song_select_index = self._song_select_index - 1
	elseif self._input.down.pressed then
		self._song_select_index = self._song_select_index + 1
	end

	if self._song_select_index < 1 then
		self._song_select_index = 1
	elseif self._song_select_index > #self._all_song_data then
		self._song_select_index = #self._all_song_data
	end

	self._karaoke_menu_panel:set_selection(self._song_select_index)

	if self._input.accept.pressed then
		self:start_song(t)

		self._karaoke_audio_panel:set_audio_movie("movies/karaoke_sfx/menu_decide", false, 3)
		self._karaoke_audio_panel:play(3)
	elseif self._input.cancel.pressed then
		self:close()
	end
end

function KaraokeManager:_update_score(t, dt)
	local time_since_start = t - self._score_start_time
	if time_since_start <= 2 and not self._prescore_state then
		self._prescore_state = true
	elseif time_since_start > 2 and self._prescore_state then 
		self._karaoke_score_panel:show()
		self._karaoke_audio_panel:pause_all()

		self._karaoke_audio_panel:set_audio_movie("movies/karaoke_sfx/result_count_loop", false, 3)
		self._karaoke_audio_panel:play(3)

		self._prescore_state = false
	end

	if self._prescore_state then return end

	if time_since_start <= 4.5 and not self._show_score_state then
		self._show_score_state = true

		self._karaoke_score_panel:jumble_numbers()
	elseif time_since_start > 4.5 and self._show_score_state then
		self._show_score_state = false

		self._karaoke_audio_panel:set_audio_movie("movies/karaoke_sfx/result_count_finish", false, 3)
		self._karaoke_audio_panel:play(3)

		local great_score = (self._score_data.great / #self._current_song_data.notes) * 100
		local good_score = (self._score_data.good / #self._current_song_data.notes) * 50

		local score = great_score + good_score

		local win_sound = "movies/karaoke_sfx/jingle_result_high"
		if score == 100 then
			win_sound = "movies/karaoke_sfx/jingle_result_perfect"
		elseif score < 50 then
			win_sound = "movies/karaoke_sfx/jingle_result_low"
		end

		local kappa = "great"
		if score < 50 then
			kappa = "failed"
		elseif score < 75 then
			kappa = "good"
		end

		self._karaoke_audio_panel:set_audio_movie(win_sound, false, 4)
		self._karaoke_audio_panel:play(4)

		self._karaoke_score_panel:set_whole_number(score)
		self._karaoke_score_panel:set_kappa(kappa)
	end

	if self._show_score_state then return end

	if self._input.accept.pressed or self._input.cancel.pressed or self._controller:get_any_input_pressed() then
		self:close()
	end
end

function KaraokeManager:show_workspace()
	self._fullscreen_workspace:show()
	self._saferect_workspace:show()
end

function KaraokeManager:hide_workspace()
	self._fullscreen_workspace:hide()
	self._saferect_workspace:hide()
end

Hooks:PostHook(GameSetup, "init_managers", "InitKaraokeManager", function(self)
	managers.karaoke = KaraokeManager:new()
end)

Hooks:PostHook(GameSetup, "update", "UpdateKaraokeManager", function(self, t, dt)
	managers.karaoke:update(t, dt)
end)

Hooks:PostHook(GameSetup, "paused_update", "PausedUpdateKaraokeManager", function(self, t, dt)
	managers.karaoke:paused_update(t, dt)
end)