KaraokeAudioGui = KaraokeAudioGui or class()

function KaraokeAudioGui:init(main_panel, data, x, y, w, h)
	if main_panel:child("karaoke_audio_panel") then
		main_panel:remove(main_panel:child("karaoke_audio_panel"))
	end

	self._karaoke_audio_panel = main_panel:panel({
		x = x,
		y = y,
		w = w,
		h = h,
		name = "karaoke_audio_panel",
		visible = true,
		layer = 0
	})

	self._speed = 1

	self._audio_videos = {}
end

function KaraokeAudioGui:set_audio_movie(path, loop, layer)
	loop = loop or false
	layer = layer or 1

	local audio_video = self._audio_videos[layer]
	if audio_video then
		managers.video:remove_video(audio_video)

		if alive(audio_video) then
			self._karaoke_audio_panel:remove(audio_video)
		end
	end

	audio_video = self._karaoke_audio_panel:video({
		video = path,
		loop = loop,
		in_mem = true,
		w = 0,
		h = 0
	})

	managers.video:add_video(audio_video)
	audio_video:pause()

	self._audio_videos[layer] = audio_video
end

function KaraokeAudioGui:play(layer)
	layer = layer or 1

	if self._audio_videos[layer] then
		self._audio_videos[layer]:play()
	end
end

function KaraokeAudioGui:pause(layer)
	layer = layer or 1

	if self._audio_videos[layer] then
		self._audio_videos[layer]:pause()
	end
end

function KaraokeAudioGui:goto_time(layer, time)
	layer = layer or 1

	if self._audio_videos[layer] then
		self._audio_videos[layer]:goto_time(time)
	end
end

function KaraokeAudioGui:play_all()
	for _, audio_video in pairs(self._audio_videos) do
		audio_video:play()
	end
end

function KaraokeAudioGui:pause_all()
	for _, audio_video in pairs(self._audio_videos) do
		audio_video:pause()
	end
end

function KaraokeAudioGui:audio_time(layer)
	layer = layer or 1

	if self._audio_videos[layer] then
		return self._audio_videos[layer]:length() * (self._audio_videos[layer]:current_frame() / self._audio_videos[layer]:frames())
	end

	return 0
end

function KaraokeAudioGui:length(layer)
	if self._audio_videos[layer] then
		return self._audio_videos[layer]:length()
	end

	return 0
end

function KaraokeAudioGui:set_volume_gain(layer, volume)
	layer = layer or 1

	if self._audio_videos[layer] then
		self._audio_videos[layer]:set_volume_gain(volume)
	end
end

function KaraokeAudioGui:all_done()
	local not_done = false
	for _, audio_video in pairs(self._audio_videos) do
		if audio_video:loop_count() < 1 then
			not_done = true
			break
		end
	end

	return not not_done
end

-- function KaraokeAudioGui:resync_to_layer(layer)
-- 	layer = layer or 1

-- 	local audio_time = self:audio_time(layer)


-- end

function KaraokeAudioGui:show()
	self._karaoke_audio_panel:set_visible(true)
end

function KaraokeAudioGui:hide()
	self._karaoke_audio_panel:set_visible(false)
end