KaraokeGui.Helpers = KaraokeGui.Helpers or {}
KaraokeGui.Helpers.GenericBar = KaraokeGui.Helpers.GenericBar or class(ExtendedPanel)

local GenericBar = KaraokeGui.Helpers.GenericBar

function GenericBar:init(parent, config)
	GenericBar.super.init(self, parent, config)

	local bar_end_texture_width = 6
	local bar_end_texture_height = 11
	local bar_end_aspect_ratio = bar_end_texture_width/bar_end_texture_height

	local bar_end_width = self._panel:height() * bar_end_aspect_ratio

	local start_texture_rect = config.start_texture_rect or { 0, 0, 1, 1 }
	local middle_texture_rect = config.middle_texture_rect or { 0, 0, 1, 1 }
	local end_texture_rect = config.end_texture_rect or { 0, 0, 1, 1 }

	self._vertical = config.vertical or false

	self._start_texture_width = math.abs(start_texture_rect[3])
	self._start_texture_height = math.abs(start_texture_rect[4])
	self._start = self._panel:bitmap({
		x = 0,
		y = 0,
		texture = config.texture or "guis/karaoke/game_ui",
		texture_rect = start_texture_rect,
		color = config.color
	})

	self._middle = self._panel:bitmap({
		x = 0,
		y = 0,
		texture = config.texture or "guis/karaoke/game_ui",
		texture_rect = middle_texture_rect,
		color = config.color
	})

	self._end_texture_width = math.abs(start_texture_rect[3])
	self._end_texture_height = math.abs(start_texture_rect[4])
	self._end = self._panel:bitmap({
		x = 0,
		y = 0,
		texture = config.texture or "guis/karaoke/game_ui",
		texture_rect = end_texture_rect,
		color = config.color
	})

	self:_update_bar_size(self._panel:width(), self._panel:height())
end

function GenericBar:_update_bar_size(width, height)
	local bar_start_aspect_ratio = self._start_texture_width/self._start_texture_height
	local bar_end_aspect_ratio = self._end_texture_width/self._end_texture_height

	if self._vertical then
		local bar_start_height = width / bar_start_aspect_ratio
		local bar_end_height = width / bar_end_aspect_ratio

		local bar_middle_height = height - (bar_start_height + bar_end_height)
		if bar_middle_height < 0 then
			local bar_ends_ratio = bar_start_height / bar_end_height

			bar_start_height = height * bar_ends_ratio
			bar_end_height = height * (1 - bar_ends_ratio)

			bar_middle_height = 0
		end

		self._start:set_top(0)
		self._start:set_width(width)
		self._start:set_height(bar_start_height)

		self._middle:set_top(bar_start_height)
		self._middle:set_width(width)
		self._middle:set_height(bar_middle_height)

		self._end:set_top(height - bar_end_height)
		self._end:set_width(width)
		self._end:set_height(bar_end_height)
	else
		local bar_start_width = height * bar_start_aspect_ratio
		local bar_end_width = height * bar_end_aspect_ratio

		local bar_middle_width = width - (bar_start_width + bar_end_width)
		if bar_middle_width < 0 then
			local bar_ends_ratio = bar_start_width / bar_end_width

			bar_start_width = width * bar_ends_ratio
			bar_end_width = width * (1 - bar_ends_ratio)

			bar_middle_width = 0
		end

		self._start:set_left(0)
		self._start:set_width(bar_start_width)
		self._start:set_height(height)

		self._middle:set_left(bar_start_width)
		self._middle:set_width(bar_middle_width)
		self._middle:set_height(height)

		self._end:set_left(width - bar_end_width)
		self._end:set_width(bar_end_width)
		self._end:set_height(height)
	end
end

local size_update_methods = {
	"set_width",
	"set_shape",
	"set_height",
	"grow",
	"set_size"
}
for _, size_update_method_name in pairs(size_update_methods) do
	GenericBar[size_update_method_name] = function(self, ...)
		self._panel[size_update_method_name](self._panel, ...)

		self:_update_bar_size(self._panel:width(), self._panel:height())
	end
end