KaraokeGui = KaraokeGui or class()

function KaraokeGui:init(main_panel, data, x, y, w, h)
	if main_panel:child("karaoke_panel") then
		main_panel:remove(main_panel:child("karaoke_panel"))
	end

	self._karaoke_panel = main_panel:panel({
		x = x,
		y = y,
		w = w,
		h = h,
		name = "karaoke_panel",
		visible = true,
		layer = 10
	})

	-- self._karaoke_panel:rect({
	-- 	x = 0,
	-- 	y = 0,
	-- 	w = self._karaoke_panel:width(),
	-- 	h = self._karaoke_panel:height(),
	-- 	color = Color.black,
	-- 	visible = true
	-- })

	self._note_element_lists_by_index = {}
	self._note_index_by_bar = {}

	self._timer_bars = {}
	self._timer_bar_lyrics = {}

	local timer_bar_start_height = 0.235
	local timer_bar_offset = 0.22
	local timer_bar_count = 4

	for index = 1, timer_bar_count do
		self._note_index_by_bar[index] = {}

		local timer_bar_x = w * 0.06
		local timer_bar_y = h * (timer_bar_start_height + (timer_bar_offset * (index - 1)))

		local timer_bar_width = self._karaoke_panel:width()
		local timer_bar_height = self._karaoke_panel:height() * 0.015

		self._timer_bars[index] = KaraokeGui.Helpers.TimerBar:new(self._karaoke_panel, {
			x = timer_bar_x,
			y = timer_bar_y,
			w = timer_bar_width,
			h = timer_bar_height,
			visible = false
		})

		local timer_bar_lyric_x = timer_bar_x
		local timer_bar_lyric_y = timer_bar_y - (self._karaoke_panel:height() * 0.18)

		local timer_bar_lyric_height = self._karaoke_panel:height() * 0.16
		local timer_bar_lyric_width = timer_bar_lyric_height * (512/64)

		self._timer_bar_lyrics[index] = KaraokeGui.Helpers.TimerBarLyric:new(self._karaoke_panel, {
			x = timer_bar_lyric_x,
			y = timer_bar_lyric_y,
			w = timer_bar_lyric_width,
			h = timer_bar_lyric_height,
			visible = false
		})
	end

	self._timer_bar_marker = KaraokeGui.Helpers.TimerBarMarker:new(self._karaoke_panel, {
		x = 0,
		y = 0,
		w = self._karaoke_panel:width() * 0.04,
		h = self._karaoke_panel:height() * 0.23,
		layer = 20,
		visible = false
	})



	-- self._test_button = KaraokeGui.Helpers.TimerBarNote:new(self._karaoke_panel, {
	-- 	x = 0,
	-- 	y = 0,
	-- 	w = self._karaoke_panel:height() * 0.25,
	-- 	h = self._karaoke_panel:height() * 0.25,
	-- 	button = "cross",
	-- 	platform = "pc"
	-- })
end

function KaraokeGui:hide_timer_bar(line)
	self._timer_bars[line]:set_visible(false)
	self._timer_bar_lyrics[line]:set_visible(false)

	self:check_marker_visibility()

	for _, index in pairs(self._note_index_by_bar[line]) do
		local note_elements = self._note_element_lists_by_index[index]
		for _, note_element in pairs(note_elements) do
			note_element:remove_self()
		end

		self._note_element_lists_by_index[index] = nil
	end

	self._note_index_by_bar[line] = {}
end

function KaraokeGui:set_timer_bar(line, lyric_texture, lyric_width)
	local first_new = not self:any_bars_visible()

	self:hide_timer_bar(line)

	self._timer_bars[line]:set_width(self._timer_bar_lyrics[line]:width() * lyric_width)
	self._timer_bars[line]:set_visible(true)

	if lyric_texture then
		self._timer_bar_lyrics[line]:set_lyric_texture(lyric_texture, lyric_width)
		self._timer_bar_lyrics[line]:set_visible(true)
	else
		self._timer_bar_lyrics[line]:set_visible(false)
	end

	if first_new then
		self:set_marker_bar(line)
	end

	self:check_marker_visibility()
end

function KaraokeGui:set_timer_bar_fill(line, fill_percentage)
	self._timer_bars[line]:set_fill(fill_percentage)
	self._timer_bar_lyrics[line]:set_fill(fill_percentage)

	if self._timer_bar_marker_line == line then
		local timer_bar = self._timer_bars[line]

		self._timer_bar_marker:set_center_x(timer_bar:x() + (timer_bar:width() * fill_percentage))
	end
end

function KaraokeGui:any_bars_visible()
	local any_visible = false
	for _, timer_bar in pairs(self._timer_bars) do
		if timer_bar:visible() then
			any_visible = true
			break
		end
	end

	return any_visible
end

function KaraokeGui:check_marker_visibility()
	self._timer_bar_marker:set_visible(self:any_bars_visible())
end

function KaraokeGui:set_marker_bar(line)
	self._timer_bar_marker_line = line
	self._timer_bar_marker:set_bottom(self._timer_bars[line]:bottom())
end

function KaraokeGui:add_note(index, platform, button, type, line, line_percentage, line_percentage_end)
	local timer_bar = self._timer_bars[line]
	local note = KaraokeGui.Helpers.TimerBarNote:new(self._karaoke_panel, {
		w = self._karaoke_panel:height() * 0.1,
		h = self._karaoke_panel:height() * 0.1,
		platform = platform,
		button = button,
		layer = 31
	})

	local center_x = timer_bar:x() + (timer_bar:width() * line_percentage)
	local center_y = timer_bar:center_y()

	note:set_center(center_x, center_y)

	self._note_element_lists_by_index[index] = {}
	table.insert(self._note_element_lists_by_index[index], note)
	table.insert(self._note_index_by_bar[line], index)

	if type == "hold" or type == "rapid_press" then
		local second_center_x = timer_bar:x() + (timer_bar:width() * line_percentage_end)

		local bar = KaraokeGui.Helpers.TimerBarHoldRapid:new(self._karaoke_panel, {
			w = second_center_x - center_x,
			h = self._karaoke_panel:height() * 0.05,
			x = center_x,
			layer = 30,
			type = type
		})
		bar:set_center_y(center_y)

		table.insert(self._note_element_lists_by_index[index], bar)

		local second_note = KaraokeGui.Helpers.TimerBarNote:new(self._karaoke_panel, {
			w = self._karaoke_panel:height() * 0.1,
			h = self._karaoke_panel:height() * 0.1,
			platform = platform,
			button = button,
			layer = 31
		})

		second_note:set_center(second_center_x, center_y)

		table.insert(self._note_element_lists_by_index[index], second_note)
	end
end

function KaraokeGui:hit_note(index)
	if self._note_element_lists_by_index[index] then
		for _, note_element in pairs(self._note_element_lists_by_index[index]) do
			note_element:hit()
		end
	end
end

function KaraokeGui:show()
	self._karaoke_panel:set_visible(true)
end

function KaraokeGui:hide()
	self._karaoke_panel:set_visible(false)
end








function KaraokeGui:add_debug_to_line(line, start_time, end_time)
	--self:add_debug_time(self._timer_bars[line], "left", tweak_data.menu.pd2_medium_font_size * 0.8, start_time)
	--self:add_debug_time(self._timer_bars[line], "right", tweak_data.menu.pd2_medium_font_size * 0.8, end_time)
end

function KaraokeGui:add_debug_to_note(index, time)
	--self:add_debug_time(self._note_element_lists_by_index[index][1], "bottom", tweak_data.menu.pd2_medium_font_size * 0.4, time)
end

function KaraokeGui:add_debug_time(element, side, size, value)
	-- self._debug_times = self._debug_times or {}

	-- local debug_time = KaraokeGui.Helpers.DebugTime:new(self._karaoke_panel, {
	-- 	track = element,
	-- 	text = tostring(math.round(value * 1000)),
	-- 	size = size
	-- })

	-- if side == "left" then
	-- 	debug_time:set_right(element:left())
	-- 	debug_time:set_center_y(element:center_y())
	-- elseif side == "right" then
	-- 	debug_time:set_left(element:right())
	-- 	debug_time:set_center_y(element:center_y())
	-- elseif side == "bottom" then
	-- 	debug_time:set_top(element:bottom())
	-- 	debug_time:set_center_x(element:center_x())
	-- elseif side == "top" then
	-- 	debug_time:set_bottom(element:top())
	-- 	debug_time:set_center_x(element:center_x())
	-- end

	-- table.insert(self._debug_times, debug_time)
end

function KaraokeGui:check_debug_times()
	-- if not self._debug_times then return end

	-- for _, debug_time in pairs(self._debug_times) do
	-- 	debug_time:check()
	-- end
end

function KaraokeGui:debug_timer(time)
	-- if not self._debug_timer then
	-- 	self._debug_timer = KaraokeGui.Helpers.DebugTime:new(self._karaoke_panel, {
	-- 		x = 0,
	-- 		y = 0,
	-- 		text = tostring(math.round(time * 1000)),
	-- 		size = tweak_data.menu.pd2_medium_font_size * 1.5
	-- 	})
	-- end

	-- self._debug_timer:set_text(tostring(math.round(time * 1000)))
end