KaraokeMenuGui = KaraokeMenuGui or class()

function KaraokeMenuGui:init(main_panel, data, x, y, w, h)
	if main_panel:child("karaoke_menu_panel") then
		main_panel:remove(main_panel:child("karaoke_menu_panel"))
	end

	self._karaoke_menu_panel = main_panel:panel({
		x = x,
		y = y,
		w = w,
		h = h,
		name = "karaoke_menu_panel",
		visible = true,
		layer = 0
	})

	self._info_text = self._karaoke_menu_panel:text({
		w = self._karaoke_menu_panel:width(),
		h = self._karaoke_menu_panel:height(),
		x = 0,
		y = 0,
		text = "( Enter / A / X ) = Enter | ( Escape / B / O ) = Exit | ( Movement Keys / D-Pad / Left Stick = Navigate )",
		align = "center",
		vertical = "bottom",
		font = tweak_data.menu.pd2_medium_font,
		font_size = tweak_data.menu.pd2_medium_font_size
	})

	self._disc_case = self._karaoke_menu_panel:bitmap({
		w = self._karaoke_menu_panel:width() * 0.52,
		h = self._karaoke_menu_panel:height() * 0.83,
		texture = "guis/karaoke/disc_case",
		layer = 2
	})
	self._disc_case:set_center(self._karaoke_menu_panel:center_x(), self._karaoke_menu_panel:center_y())

	self._disc_panel = self._karaoke_menu_panel:panel({
		x = self._disc_case:right() - (self._karaoke_menu_panel:width() * 0.02),
		w = self._karaoke_menu_panel:width() * 0.18,
		h = self._karaoke_menu_panel:height() * 0.70
	})
	self._disc_panel:set_center_y(self._karaoke_menu_panel:center_y())

	self._disc_sticker = self._disc_panel:bitmap({
		x = 0,
		y = 0,
		w = self._disc_panel:width(),
		h = self._disc_panel:height(),
		layer = 1,
		texture = "guis/karaoke/disc_sticker/ontheroad"
	})

	self._disc = self._disc_panel:bitmap({
		x = 0,
		y = 0,
		w = self._disc_panel:width(),
		h = self._disc_panel:height(),
		texture = "guis/karaoke/disc"
	})

	self._song_select_bg = self._karaoke_menu_panel:rect({
		w = self._disc_case:width() * 0.9,
		h = self._disc_case:height() * 0.43,
		color = Color(255, 53, 44, 40) / 255,
		layer = 3
	})
	self._song_select_bg:set_center(self._disc_case:center_x(), self._disc_case:center_y())

	self._song_select_top_text = self._karaoke_menu_panel:text({
		w = self._song_select_bg:width(),
		h = self._song_select_bg:height() * 0.1,
		x = self._song_select_bg:left(),
		y = self._song_select_bg:top(),
		text = "Song Selection",
		align = "center",
		vertical = "center",
		font = tweak_data.menu.pd2_medium_font,
		font_size = tweak_data.menu.pd2_medium_font_size * 0.7,
		layer = 4
	})

	self._song_sticker_lookup = {}

	self._song_select_song_indicator = {}
	self._song_select_song_bgs = {}
	self._song_select_song_titles = {}

	local song_y = self._song_select_bg:top() + (self._song_select_bg:height() * 0.1)
	local song_spacer = self._song_select_bg:height() * 0.01

	local song_limit = 5
	for index = 1, song_limit do
		local song_bg = self._karaoke_menu_panel:rect({
			w = self._song_select_bg:width() * 0.98,
			h = self._song_select_bg:height() * 0.17,
			y = song_y,
			color = Color(255, 214, 185, 195) / 255,
			layer = 6
		})
		song_bg:set_center_x(self._disc_case:center_x())
		song_y = song_y + song_bg:height() + song_spacer

		local song_indicator = self._karaoke_menu_panel:rect({
			w = song_bg:width() + 8,
			h = song_bg:height() + 8,
			color = Color(255, 212, 96, 190) / 255,
			layer = 5,
			visible = false
		})
		song_indicator:set_center(song_bg:center_x(), song_bg:center_y())

		local song_title = self._karaoke_menu_panel:text({
			w = song_bg:width(),
			h = song_bg:height(),
			x = song_bg:left(),
			y = song_bg:top(),
			text = "",
			align = "center",
			vertical = "center",
			font = tweak_data.menu.pd2_medium_font,
			font_size = tweak_data.menu.pd2_medium_font_size,
			color = Color.black,
			layer = 7
		})

		self._song_select_song_indicator[index] = song_indicator
		self._song_select_song_bgs[index] = song_bg
		self._song_select_song_titles[index] = song_title
	end
end

function KaraokeMenuGui:set_song_info(index, name, sticker)
	self._song_select_song_titles[index]:set_text(name)
	self._song_sticker_lookup[index] = sticker
end

function KaraokeMenuGui:set_selection(index)
	if index == self._last_index then return end
	self._last_index = index

	for _, indicator in pairs(self._song_select_song_indicator) do
		indicator:set_visible(false)
	end

	self._song_select_song_indicator[index]:set_visible(true)
	self._disc_sticker:set_image(self._song_sticker_lookup[index])

	function animate_disc(o)
		local start_left = self._disc_case:right() - self._disc_panel:width()
		local target_left = self._disc_case:right() - (self._karaoke_menu_panel:width() * 0.02)

		local time = 0.3
		local lerp_speed = 1 / time

		local t, dt = 0
		local lerp = 0

		local left_lerped = start_left

		while true do
			dt = coroutine.yield()
			t = t + dt

			lerp = math.min(t * lerp_speed, 1)
			left_lerped = math.lerp(start_left, target_left, lerp)
			self._disc_panel:set_left(left_lerped)

			if lerp >= 1 then return end
		end
	end
	self._disc_panel:animate(animate_disc)
end

function KaraokeMenuGui:show()
	self._karaoke_menu_panel:set_visible(true)
end

function KaraokeMenuGui:hide()
	self._karaoke_menu_panel:set_visible(false)
end