KaraokeScoreGui = KaraokeScoreGui or class()

function KaraokeScoreGui:init(main_panel, data, x, y, w, h)
	if main_panel:child("karaoke_score_panel") then
		main_panel:remove(main_panel:child("karaoke_score_panel"))
	end

	self._karaoke_score_panel = main_panel:panel({
		x = x,
		y = y,
		w = w,
		h = h,
		name = "karaoke_score_panel",
		visible = true,
		layer = 0
	})

	self._background = self._karaoke_score_panel:bitmap({
		w = self._karaoke_score_panel:width() * 1333/1797,
		h = self._karaoke_score_panel:height() * 739/1011,
		texture = "guis/karaoke/score_bg",
		layer = 1
	})
	self._background:set_center(self._karaoke_score_panel:center_x(), self._karaoke_score_panel:center_y())

	self._kappa = self._karaoke_score_panel:bitmap({
		w = self._background:width() * 795/888,
		h = self._background:height() * 202/492,
		y = self._background:top() + (self._background:height() * 247/492),
		texture = "guis/karaoke/score_kappa",
		layer = 2
	})
	self._kappa:set_center_x(self._background:center_x())

	self._numbers = {}
	local number_count = 3

	for index = 1, number_count do
		self._numbers[index] = self._karaoke_score_panel:bitmap({
			w = self._background:width() * 80/888,
			h = self._background:height() * 106/492,
			x = self._background:left() + (self._background:width() * 254/888) + (self._background:width() * 148/888 * (index - 1)),
			y = self._background:top() + (self._background:height() * 110/492),
			texture = "guis/karaoke/score_numbers",
			layer = 2
		})
	end
end

function KaraokeScoreGui:blank_numbers()
	self:set_number(1, nil)
	self:set_number(2, nil)
	self:set_number(3, nil)
end

function KaraokeScoreGui:set_whole_number(number)
	local first_number = math.floor(number / 100)
	if first_number <= 0 then
		first_number = nil
	else
		first_number = first_number % 10
	end

	local second_number = math.floor(number / 10)
	if second_number <= 0 then
		second_number = nil
	else
		second_number = second_number % 10
	end

	local third_number = math.floor(number) % 10

	self:set_number(1, first_number)
	self:set_number(2, second_number)
	self:set_number(3, third_number)
end

function KaraokeScoreGui:set_number(index, number, automated)
	if not automated then
		self._numbers[index]:stop()
	end

	local width = 80
	local height = 106

	local offset = 81

	if number == nil or number == false then
		number = -1
	end

	self._numbers[index]:set_texture_rect(
		offset * (number + 1),
		0,
		width,
		height
	)
end

function KaraokeScoreGui:jumble_numbers()
	self:jumble_number(1)
	self:jumble_number(2)
	self:jumble_number(3)
end

function KaraokeScoreGui:jumble_number(index)
	self._numbers[index]:stop()

	function animate_number(o)
		local frame_length = 0.05
		local t, dt = 0

		while true do
			dt = coroutine.yield()
			t = t + dt

			if t > frame_length then
				t = 0

				self:set_number(index, math.round(math.random() * 9), true)
			end
		end
	end

	self._numbers[index]:animate(animate_number)
end

function KaraokeScoreGui:set_kappa(type)
	self._kappa:stop()

	if not type then		
		self._kappa:set_visible(false)
		return
	end

	self._kappa:set_visible(true)

	function animate_kappa(o)
		local x_offsets = {
			great = 0,
			good = 795,
			failed = 1590
		}

		local x_offset = x_offsets[type or "great"]

		local frame_length = 0.25
		local t, dt = 0

		local on_frame_one = true
		self._kappa:set_texture_rect(x_offset, 0, 795, 202)

		while true do
			dt = coroutine.yield()
			t = t + dt

			if t > frame_length then
				t = 0

				if on_frame_one then
					on_frame_one = false
					self._kappa:set_texture_rect(x_offset, 202, 795, 202)
				else
					on_frame_one = true
					self._kappa:set_texture_rect(x_offset, 0, 795, 202)
				end
			end
		end
	end

	self._kappa:animate(animate_kappa)
end

function KaraokeScoreGui:show()
	self._karaoke_score_panel:set_visible(true)

	self:set_kappa(nil)
	self:blank_numbers()
end

function KaraokeScoreGui:hide()
	self._karaoke_score_panel:set_visible(false)

	self:set_kappa(nil)
	self:blank_numbers()
end