GameCutsceneCamera = GameCutsceneCamera or class()

function GameCutsceneCamera:init(unit)
	self._unit = unit

	self._camera_target_object = self._unit:get_object(Idstring("a_camera_target"))
	self._camera_controller_object = self._unit:get_object(Idstring("a_camera_controller"))

	self._camera = World:create_camera()

	self._default_fov = 75

	self._camera:set_fov(self._default_fov)
	self._camera:set_near_range(1)
	self._camera:set_far_range(250000)

	self._viewport = managers.viewport:new_vp(0, 0, 1, 1, "GameCutsceneCamera", 20 + self._priority)
	self._director = self._viewport:director()
	self._shaker = self._director:shaker()
	self._camera_controller = self._director:make_camera(self._camera, Idstring("previs_camera"))

	self._viewport:set_camera(self._camera)
	self._director:set_camera(self._camera_controller)
	self._director:position_as(self._camera)
	self._camera_controller:set_both(self._camera_target_object)

	self._viewport:set_active(true)

	self._debug = false
end

function GameCutsceneCamera:update(unit, t, dt)
	local camera_controller_vector = self._camera_controller_object:local_position()

	self._camera:set_fov(camera_controller_vector.x)
	-- managers.environment_controller:set_dof_setting("standard")
	-- managers.environment_controller:set_dof_distance(camera_controller_vector.y, true)

	if self._debug then
		local position = self._camera_target_object:position()
		local rotation = self._camera_target_object:rotation()

		Application:draw_sphere(position, 5, 1, 0, 0)
		Application:draw_line(position, position + (rotation:y() * 10), 1, 0, 0)
	end
end

function GameCutsceneCamera:destroy()
	if self._viewport then
		self._viewport:destroy()

		self._viewport = nil
	end

	if alive(self._camera) then
		World:delete_camera(self._camera)

		self._camera = nil
	end
end